# --- Day 3: Crossed Wires ---

# The gravity assist was successful, and you're well on your way to the Venus
# refuelling station. During the rush back on Earth, the fuel management system
# wasn't completely installed, so that's next on the priority list.

# Opening the front panel reveals a jumble of wires. Specifically, two wires
# are connected to a central port and extend outward on a grid. You trace the
# path each wire takes as it leaves the central port, one wire per line of text
# (your puzzle input).

# The wires twist and turn, but the two wires occasionally cross paths. To fix
# the circuit, you need to find the intersection point closest to the central
# port. Because the wires are on a grid, use the Manhattan distance for this
# measurement. While the wires do technically cross right at the central port
# where they both start, this point does not count, nor does a wire count as
# crossing with itself.

# For example, if the first wire's path is R8,U5,L5,D3, then starting from the
# central port (o), it goes right 8, up 5, left 5, and finally down 3:

# ...........
# ...........
# ...........
# ....+----+.
# ....|....|.
# ....|....|.
# ....|....|.
# .........|.
# .o-------+.
# ...........

# Then, if the second wire's path is U7,R6,D4,L4, it goes up 7, right 6, down
# 4, and left 4:

# ...........
# .+-----+...
# .|.....|...
# .|..+--X-+.
# .|..|..|.|.
# .|.-X--+.|.
# .|..|....|.
# .|.......|.
# .o-------+.
# ...........

# These wires cross at two locations (marked X), but the lower-left one is
# closer to the central port: its distance is 3 + 3 = 6.

# Here are a few more examples:

#     R75,D30,R83,U83,L12,D49,R71,U7,L72
#     U62,R66,U55,R34,D71,R55,D58,R83 = distance 159
#     R98,U47,R26,D63,R33,U87,L62,D20,R33,U53,R51
#     U98,R91,D20,R16,D67,R40,U7,R15,U6,R7 = distance 135

# What is the Manhattan distance from the central port to the closest
# intersection?

from math import inf
from typing import List

from sympy import Segment, Point

with open("files/P3.txt") as f:
    # with open("files/test3.txt") as f:
    wire_a, wire_b = [path.split(",") for path in f.read().strip().split()]


def get_final_point(init: List[int], direction: str, length: int) -> List[int]:
    if direction == "R":
        init[0] += length
    elif direction == "L":
        init[0] -= length
    elif direction == "U":
        init[1] += length
    elif direction == "D":
        init[1] -= length
    return init


def get_path(wire: List[str]) -> List[List[int]]:
    init = [0, 0]
    path = []
    for step in wire:
        direction, length = step[0], int(step[1:])
        final_point = get_final_point(init, direction, length)
        # make a copy of the final point to avoid reference to same object
        path.append(final_point.copy())
        init = final_point
    return path


path_a = get_path(wire_a)
path_b = get_path(wire_b)


def part_1() -> None:
    closest = inf
    for p1, p2 in zip(path_a[:-1], path_a[1:]):
        s1 = Segment(p1, p2)
        for p3, p4 in zip(path_b[:-1], path_b[1:]):
            s2 = Segment(p3, p4)
            intersect = s1.intersection(s2)
            if intersect:
                manhattan_distance = abs(intersect[0][0]) + abs(
                    intersect[0][1]
                )
                if manhattan_distance < closest:
                    closest = manhattan_distance

    print(f"The manhattan distance to the closest intersection is {closest}")


# --- Part Two ---

# It turns out that this circuit is very timing-sensitive; you actually need to
# minimize the signal delay.

# To do this, calculate the number of steps each wire takes to reach each
# intersection; choose the intersection where the sum of both wires' steps is
# lowest. If a wire visits a position on the grid multiple times, use the steps
# value from the first time it visits that position when calculating the total
# value of a specific intersection.

# The number of steps a wire takes is the total number of grid squares the wire
# has entered to get to that location, including the intersection being
# considered. Again consider the example from above:

# ...........
# .+-----+...
# .|.....|...
# .|..+--X-+.
# .|..|..|.|.
# .|.-X--+.|.
# .|..|....|.
# .|.......|.
# .o-------+.
# ...........

# In the above example, the intersection closest to the central port is reached
# after 8+5+5+2 = 20 steps by the first wire and 7+6+4+3 = 20 steps by the
# second wire for a total of 20+20 = 40 steps.

# However, the top-right intersection is better: the first wire takes only
# 8+5+2 = 15 and the second wire takes only 7+6+2 = 15, a total of 15+15 = 30
# steps.

# Here are the best steps for the extra examples from above:

#     R75,D30,R83,U83,L12,D49,R71,U7,L72
#     U62,R66,U55,R34,D71,R55,D58,R83 = 610 steps
#     R98,U47,R26,D63,R33,U87,L62,D20,R33,U53,R51
#     U98,R91,D20,R16,D67,R40,U7,R15,U6,R7 = 410 steps

# What is the fewest combined steps the wires must take to reach an
# intersection?


def accumulate_path(wire, steps):
    path = 0
    for step in wire[:steps]:
        _, length = step[0], int(step[1:])
        path += length
    return path


def get_direction(segment):
    is_horizontal = False
    if segment.points[0][1] == segment.points[1][1]:
        is_horizontal = True

    if is_horizontal:
        if segment.points[0][0] > segment.points[1][0]:
            direction = "L"
        else:
            direction = "R"
    else:
        if segment.points[0][1] > segment.points[1][1]:
            direction = "D"
        else:
            direction = "U"

    return is_horizontal, direction


def get_range(segment, is_horizontal, direction):
    if is_horizontal:
        if direction == "L":
            return range(segment.points[0][0], segment.points[1][0], -1)
        else:
            return range(segment.points[0][0], segment.points[1][0])
    else:
        if direction == "D":
            return range(segment.points[0][1], segment.points[1][1], -1)
        else:
            return range(segment.points[0][1], segment.points[1][1])


def find_intersection(segment_a, segment_b, intersection_point):
    is_horizontal_a, direction_a = get_direction(segment_a)
    is_horizontal_b, direction_b = get_direction(segment_b)

    range_a = get_range(segment_a, is_horizontal_a, direction_a)
    range_b = get_range(segment_b, is_horizontal_b, direction_b)

    steps = 0
    for idx_a, a in enumerate(range_a):
        for idx_b, b in enumerate(range_b):
            if (
                Point(a, b) == intersection_point
                or Point(b, a) == intersection_point
            ):
                steps = idx_a + idx_b
    return steps


def part_2() -> None:
    closest = inf
    for idx_a, (p1, p2) in enumerate(zip(path_a[:-1], path_a[1:]), start=1):
        s1 = Segment(p1, p2)
        for idx_b, (p3, p4) in enumerate(
            zip(path_b[:-1], path_b[1:]), start=1
        ):
            s2 = Segment(p3, p4)
            intersect = s1.intersection(s2)
            if intersect:
                until_intersection = find_intersection(s1, s2, intersect[0])
                steps_a = accumulate_path(wire_a, idx_a)
                steps_b = accumulate_path(wire_b, idx_b)
                sum_of_steps = steps_a + steps_b + until_intersection
                if sum_of_steps < closest:
                    closest = sum_of_steps

    print(f"The minimum steps to an intersection is {closest}")


if __name__ == "__main__":
    part_1()
    part_2()
